" Vim syntax file
" Language: POV-Ray v3.8 (povr fork) Scene Description Language
" Maintainer: William F. Pokorny
" URLs:
"    http://news.povray.org/povray.binaries.programming/
"    http://news.povray.org/povray.pov4.discussion.general/
" Copyright:
"    Copyright (c) 1991-2023 Persistence of Vision Raytracer Pty. Ltd.
" License:
"    This work is licensed under the Creative Commons Attribution-ShareAlike 4.0
"    International Public License.
"    Visit https://creativecommons.org/licenses/by-sa/4.0/
" Last Change: Thu Aug 31 12:51:32 -0400 2023

if exists("b:current_syntax")
    finish
endif

" TODO. Why do we see the echo below twice when uncommented?
" echom "Debugging. Reading povr.vim."

let s:cpo_save = &cpo
set cpo&vim

syn case match

" Comments
syn region povrComment start="/\*" end="\*/" contains=povrTodo,povrComment
syn match povrComment "//.*" contains=povrTodo
syn match povrCommentErrors "\*/"
syn keyword povrTodo TODO FIXME NB NOTE contained

" Active, traditional keywords
syn keyword povrKeywords aa_level aa_threshold abs absorption accuracy acos acosh adaptive adc_bailout agate
syn keyword povrKeywords albedo all altitude always_sample ambient angle anisotropy aoi aperture append arc_angle
syn keyword povrKeywords area_illumination area_light array asc ascii asin asinh assumed_gamma atan atan2 atanh
syn keyword povrKeywords autostop average b_spline background bezier_spline bicubic_patch bitwise_and bitwise_or
syn keyword povrKeywords bitwise_xor black_hole blend_gamma blend_mode blob blue blur_samples bmp bokeh bounded_by
syn keyword povrKeywords box boxed bozo brick brick_size brightness brilliance bt2020 bt709 bump_map bump_size
syn keyword povrKeywords bumps camera caustics ceil cells charset checker chr circular clipped_by clock clock_delta
syn keyword povrKeywords clock_on collect color color_map colour colour_map component concat cone confidence
syn keyword povrKeywords conic_sweep conserve_energy contained_by control0 control1 coords cos cosh count crackle
syn keyword povrKeywords crand cube cubic cubic_spline cubic_wave cutaway_textures cylinder cylindrical datetime
syn keyword povrKeywords default defined degrees density density_file density_map dents deprecated df3 dictionary
syn keyword povrKeywords difference diffuse dimension_size dimensions direction disc dispersion dispersion_samples
syn keyword povrKeywords dist_exp distance div double_illuminate eccentricity emission error_bound exp expand_thresholds
syn keyword povrKeywords exponent exr exterior extinction face_indices facets fade_color fade_colour fade_distance
syn keyword povrKeywords fade_power falloff falloff_angle false file_exists filter final_clock final_frame
syn keyword povrKeywords finish fisheye flatness flip floor focal_point fog form frame_number frequency fresnel
syn keyword povrKeywords function gamma gather gif global global_lights global_settings gradient granite gray
syn keyword povrKeywords gray_threshold green grey grey_threshold hdr height_field hexagon hierarchy hollow
syn keyword povrKeywords hypercomplex iff image_height image_map image_pattern image_width importance initial_clock
syn keyword povrKeywords initial_frame input_file_name inside inside_vector int interior interior_texture internal
syn keyword povrKeywords interpolate intersection intervals inverse ior irid irid_wavelength isosurface jitter
syn keyword povrKeywords jpeg julia julia_fractal lambda lathe lemon leopard light_group light_source linear_spline
syn keyword povrKeywords linear_sweep ln load_file local location log look_at looks_like low_error_factor magnet
syn keyword povrKeywords major_radius mandel map_type marble material material_map matrix max max_extent max_gradient
syn keyword povrKeywords max_iteration max_sample max_trace max_trace_level maximum_reuse media media_attenuation
syn keyword povrKeywords media_interaction merge mesh mesh2 mesh_camera metallic method metric min min_extent
syn keyword povrKeywords minimum_reuse mixed mm_per_unit mod mortar natural_spline nearest_count no no_bump_scale
syn keyword povrKeywords no_image no_radiosity no_reflection no_shadow noise_generator normal normal_indices
syn keyword povrKeywords normal_map normal_vectors now number_of_sides number_of_tiles object octaves off offset
syn keyword povrKeywords omega omnimax on once onion open optional orient orientation orthographic ovus panoramic
syn keyword povrKeywords parallel parametric pass_through pattern pavement perspective pgm phase phong phong_size
syn keyword povrKeywords photons pi pigment pigment_map pigment_pattern planar plane png point_at poly poly_wave
syn keyword povrKeywords polygon polynomial pot potential pow ppm precision precompute premultiplied pretrace_end
syn keyword povrKeywords pretrace_start prism prod projected_through pwr quadratic_spline quadric quartic quaternion
syn keyword povrKeywords quick_color quick_colour quilted radial radians radiosity radius rainbow ramp_wave
syn keyword povrKeywords rand ratio read reciprocal recursion_limit red reflection refraction repeat rgb rgbf
syn keyword povrKeywords rgbft rgbt right ripples rotate roughness samples save_file scale scallop_wave scattering
syn keyword povrKeywords seed select shadowless sin sine_wave sinh sint16be sint16le sint32be sint32le sint8
syn keyword povrKeywords size sky sky_sphere slice slope slope_map smooth smooth_triangle solid sor spacing
syn keyword povrKeywords specular sphere sphere_sweep spherical spiral1 spiral2 spline split_union spotlight
syn keyword povrKeywords sqr sqrt square srgb srgbf srgbft srgbt str strcmp strength strlen strlwr strupr sturm
syn keyword povrKeywords substr subsurface sum superellipsoid sys t tan tanh target tau text texture texture_list
syn keyword povrKeywords texture_map tga thickness threshold tiff tightness tiling tolerance toroidal torus
syn keyword povrKeywords trace transform translate translucency transmit triangle triangle_wave triangular
syn keyword povrKeywords true ttf turbulence type u u_steps uint16be uint16le uint8 ultra_wide_angle union
syn keyword povrKeywords unofficial up use_alpha use_color use_colour use_index user_defined utf8 uv_indices
syn keyword povrKeywords uv_mapping uv_vectors v v_steps val variance vaxis_rotate vcross vdot version vertex_vectors
syn keyword povrKeywords vlength vnormalize vrotate vstr vturbulence warp water_level waves width wood wrinkles
syn keyword povrKeywords write x xyz y yes z

" Keywords new to povr
syn keyword povrNewKeywords amplify bevy bool_object cmap dens_object e fork_str fork_val function_interval getenv
syn keyword povrNewKeywords grooved hard_object hi id_type intensity_max inv_pnormal_cure ip_count ip_fade_distance
syn keyword povrNewKeywords ip_frequency ip_phase ip_scale ip_seed it_depth it_frequency it_lambda it_octaves
syn keyword povrNewKeywords it_omega it_scale it_translate list_object lo log2 micro noise_cubed pattern_modifiers
syn keyword povrNewKeywords phong_odebla raw_wave report seedfi shadow_tolerance soft_object specular_odebla tiled
syn keyword povrNewKeywords uint32be uint32le word_get word_is

" Keywords removed from povr
syn keyword povrKeywordErrors agate_turb all_intersections alpha ambient_light composite dtag evaluate fclose fog_alt
syn keyword povrKeywordErrors fog_offset fog_type fopen hf_gray_16 hf_grey_16 max_intersections number_of_waves
syn keyword povrKeywordErrors polarity reflection_exponent sblue sgreen spotted sred tile2 tiles turb_depth

" Odd keywords to avoid as experimental compiles might use them
syn keyword povrOddwordErrors breakpoint distance_maximum dtag obj prefix suffix

" Directives
syn match povrDirectives "#\(break\|case\|debug\|declare\|else\|elseif\|end\|error\|fclose\|fopen\|for\|if\|ifdef\|ifndef\|include\)\>"
syn match povrDirectives "#\(local\|macro\|range\|read\|switch\|undef\|version\|warning\|while\|write\)\>"

" Removed directives
syn match povrDirectiveErrors "#\(render\|statistics\)\>"

" Items long not real directives
syn match povrDirectiveErrors "#\(default\|global_settings\)\>"

" Curtail spurious use of '#'. Parser, too often, still treats it as space.
" Note. The errors caught here not a complete compliment of current directives
syn match povrHashErrors "##\+"
syn match povrHashErrors "#\(\s\+\|$\)"
syn match povrHashErrors "#[A-Z,0-9,_]\+"
syn match povrHashErrors "[a-z,A-Z,0-9,_]\+#"
syn match povrHashErrors "#[a-z]\(\s\+\|$\)"
syn match povrHashErrors "#[a,g,h,j,k,n-q,t,x-z]"
syn match povrHashErrors "#[c-f,i,l-m,r-s,u-w][\s,b,d,g,i-k,m,p-q,s-v,x-z]"
syn match povrHashErrors "#[a-e,g-h,j-z]\{2\}\(\s\+\|$\)"
syn match povrHashErrors "#[a-c,g-m,p-q,s-z]\{3\}\(\s\+\|$\)"

" Active functions (ibuilt declared active via cfunctions.inc & functions.inc)
syn keyword povrFunctions cf_abs cf_acos cf_acosh cf_add cf_asin cf_asinh cf_atan cf_atanh cf_cmplx cf_conj
syn keyword povrFunctions cf_cos cf_cosh cf_div cf_exp cf_imag cf_inv cf_ln cf_log cf_mul cf_neg cf_norm cf_phase
syn keyword povrFunctions cf_polar cf_pow cf_proj cf_real cf_sin cf_sinh cf_sqr cf_sqrt cf_sub cf_tan cf_tanh
syn keyword povrFunctions f_agate f_boom f_box f_boxb f_bump f_cbrt f_clock f_cone f_cubic_wave f_cylinder f_dec2x_f32
syn keyword povrFunctions f_dec3x_f21 f_dec_nbit3c f_distribution f_dturbulence f_eblob f_ellipsoid f_elliptical_cylinder
syn keyword povrFunctions f_elliptical_sphrswp f_enc2x_f32 f_enc3x_f21 f_enc_nbit3c f_gradient f_granite23 f_hash
syn keyword povrFunctions f_helix1 f_helix2 f_hessian_dist f_hetero_mf f_hypot f_interior_fade f_labsweep f_lame
syn keyword povrFunctions f_length f_length_sqr f_marble f_max1to8pairs f_min1to8pairs f_morph2to9fncts f_mult1to8pairs
syn keyword povrFunctions f_noise3d f_noise_cubed f_normal f_normalized f_npmod f_null f_odd f_onion f_paraboloid
syn keyword povrFunctions f_ph f_planar f_polyhedron f_quartic_cylinder f_quilt_cubic f_r f_radial f_radial_tubes
syn keyword povrFunctions f_randn f_range2range f_remap f_repeat f_revolution f_ridge f_ridged_mf f_rotate2d
syn keyword povrFunctions f_rounded_box f_scallop_wave f_scallop_wavef f_sign f_sign0 f_signpow f_signsqrt f_sine_wave
syn keyword povrFunctions f_sine_wavef f_snoise3d f_sphere f_spiral f_superellipsoid f_supertorus f_th f_torus
syn keyword povrFunctions f_toupie f_triangle_wave f_turbulence f_upoly f_vangle f_wood f_wrap

" Removed functions
syn keyword povrFunctionErrors f_algbr_cyl1 f_algbr_cyl2 f_algbr_cyl3 f_algbr_cyl4 f_bicorn f_bifolia f_blob f_blob2
syn keyword povrFunctionErrors f_boy_surface f_comma f_cross_ellipsoids f_crossed_trough f_cubic_saddle f_cushion
syn keyword povrFunctionErrors f_devils_curve f_devils_curve_2d f_dupin_cyclid f_flange_cover f_folium_surface f_folium_surface_2d
syn keyword povrFunctionErrors f_glob f_heart f_helical_torus f_hex_x f_hex_y f_hunt_surface f_hyperbolic_torus f_isect_ellipsoids
syn keyword povrFunctionErrors f_kampyle_of_eudoxus f_kampyle_of_eudoxus_2d f_klein_bottle f_kummer_surface_v1 f_kummer_surface_v2
syn keyword povrFunctionErrors f_lemniscate_of_gerono f_lemniscate_of_gerono_2d f_mesh1 f_mitre f_nodal_cubic f_noise_generator
syn keyword povrFunctionErrors f_ovals_of_cassini f_parabolic_torus f_pillow f_piriform f_piriform_2d f_poly4 f_polytubes
syn keyword povrFunctionErrors f_quantum f_quartic_paraboloid f_quartic_saddle f_spikes f_spikes_2d f_steiners_roman
syn keyword povrFunctionErrors f_strophoid f_strophoid_2d f_torus2 f_torus_gumdrop f_umbrella f_witch_of_agnesi f_witch_of_agnesi_2d

" Keywords and inbuilt f_ functions previously faked in official POV-Ray releases
" by various shipped include defined functions. In the povr fork, some few were
" implemented as true inbuilts and are not considered errors, but active keywords
" and functions. Those below, without f_ prefixes, often now have F_ forms. This
" naming makes clear they are not inbuilt. For example, the old cosd, fake keyword,
" is in povr F_cosd. F_cosd() is defined in munctions.inc with the macro-function
" wrapping technique to compile the function only once.
syn keyword povrFunctionErrors acosd adj_range adj_range2 asind atan2d atand clamp clip cosd even f_boxed f_bozo
syn keyword povrFunctionErrors f_brick f_bumps f_checker f_crackle f_cylindrical f_dents f_gradientX f_gradientY
syn keyword povrFunctionErrors f_gradientZ f_granite f_hexagon f_leopard f_mandel f_ripples f_spherical f_spiral1
syn keyword povrFunctionErrors f_spiral2 f_spotted f_sqr f_waves f_wrinkles max3 min3 odd sgn sind tand

" String literals.
syn region  povrString        start=+"+ skip=+\\\\\|\\"+ end=+"+ contains=povrStringSpecial
" \u limited to base plane (4 hex chars). \U 6 hex chars, but >base plane support iffy
syn match   povrStringSpecial "\\u\x\{4\}|\\U\x\{6\}\ze\(\s\+\|\"\)" contained containedin=povrString
syn match   povrStringErrors  "\\u\x\{1,3\}\ze\(\s\+\|\"\)" contained containedin=povrString
syn match   povrStringErrors  "\\U\x\{1,5\}\ze\(\s\+\|\"\)" contained containedin=povrString
syn match   povrStringErrors  "\\u\x\{5,\}" contained containedin=povrString
syn match   povrStringErrors  "\\U\x\{7,\}" contained containedin=povrString
syn match   povrStringErrors  "\\x\x\+" contained containedin=povrString

" Catch common typos for those used to operators in other languages.
syn match povrOperatorErrors "||\+"
syn match povrOperatorErrors "&&\+"
syn match povrOperatorErrors "==\+"
syn match povrOperatorErrors "[\^]\+"
syn match povrOperatorErrors "\*[\*]\+"
syn match povrOperatorErrors "[\+][\+]\+"
syn match povrOperatorErrors "[\-][\-]\+"

" Catch some matching paren, brace, bracket errors
syn region povrParen start='(' end=')' contains=ALLBUT,povrParenError,povrTodo transparent
syn match povrParenError ")"
syn region povrBrace start='{' end='}' contains=ALLBUT,povrBraceError,povrTodo transparent
syn match povrBraceError "\}"
syn region povrBracket start='\[' end='\]' contains=ALLBUT,povrBracketError,povrTodo transparent
syn match povrBracketError "\]"

" Links to highlights. Mostly not aligning to predefined highlight categories.
" With POV-Ray, povr included, keywords often fit multiple categories. Herein 
" only aiming for generalized highlighting buckets. 
hi def link povrComment         Comment
hi def link povrTodo            Todo
hi def link povrKeywords        Type
hi def link povrNewKeywords     Statement
hi def link povrDirectives      PreProc
hi def link povrFunctions       Identifier
hi def link povrString          Constant

hi def link povrCommentErrors   Error
hi def link povrKeywordErrors   Error
hi def link povrOddwordErrors   Error
hi def link povrDirectiveErrors Error
hi def link povrHashErrors      Error
hi def link povrFunctionErrors  Error
hi def link povrOperatorErrors  Error
hi def link povrStringErrors    Error
hi def link povrParenError      Error
hi def link povrBraceError      Error
hi def link povrBracketError    Error

" TODO Is performance of fromstart OK on large files? 
syntax sync fromstart

let b:current_syntax = "povr"

let &cpo = s:cpo_save
unlet s:cpo_save

" SPDX-License-Identifier: CC-BY-SA-4.0
